/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.ignition.jmx;

import com.vii.brillien.ignition.jmx.mbeans.*;
import com.vii.brillien.kernel.BrillienException;

import javax.management.MBeanServerConnection;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import java.io.IOException;

/**
 * Connection to a running Brillien instance
 */
public class BrillienConnection {

    /**
     * Connection information
     */
    private String address, serverName;
    private int port;

    /**
     * JMX Connector
     */
    private JMXConnector jmxc;

    /**
     * MBeanServerConnection for JMX-based conneciton
     */
    private MBeanServerConnection mbsc;

    /**
     * Proxy for ThreadMXBean
     */
    private ThreadImpl thread;

    /**
     * Proxy for MemoryMXBean
     */
    private MemoryImpl memory;

    /**
     * Proxy for ClassLoadingMXBean
     */
    private ClassLoadingImpl classloading;

    /**
     * Proxy for RuntimeMXBean
     */
    private RuntimeImpl runtime;

    /**
     * Proxy for OperatingSystemMXBean
     */
    private OperatingSystemImpl operatingSystem;

    /**
     * Proxy for LiaisonMXBean
     */
    private LiaisonImpl liaison;

    /**
     * Creates a new BrillienConnection instance and try to connect to the running Brillien application server
     */
    public BrillienConnection(String address, int port, String serverName) throws BrillienException {
        this.address = address;
        this.port = port;
        this.serverName = serverName;

        try {
            init();
        } catch (Exception e) {
            e.printStackTrace();
            throw new BrillienException(e.getMessage());
        }
    }

    /**
     * Initializer method to get MXBean references
     */
    public void init() throws Exception {
        if (jmxc != null)
            finish();


        JMXServiceURL url = new JMXServiceURL(
                "service:jmx:rmi:///jndi/rmi://" + this.address + ":" + this.port + "/" + this.serverName
        );

        jmxc = JMXConnectorFactory.connect(url, null);

        mbsc = jmxc.getMBeanServerConnection();

        thread = new ThreadImpl(mbsc);
        memory = new MemoryImpl(mbsc);
        classloading = new ClassLoadingImpl(mbsc);
        runtime = new RuntimeImpl(mbsc);
        operatingSystem = new OperatingSystemImpl(mbsc);
        liaison = new LiaisonImpl(mbsc);
    }

    /**
     * Getter method for the field thread
     */
    public ThreadImpl getThread() {
        return thread;
    }

    /**
     * Getter method for the field memory
     */
    public MemoryImpl getMemory() {
        return memory;
    }

    /**
     * Getter method for the field classloading
     */
    public ClassLoadingImpl getClassloading() {
        return classloading;
    }

    /**
     * Getter method for the field runtime
     */
    public RuntimeImpl getRuntime() {
        return runtime;
    }

    /**
     * Getter method for the field operatingSystem
     */
    public OperatingSystemImpl getOperatingSystem() {
        return operatingSystem;
    }

    /**
     * Getter method for the field liaison
     */
    public LiaisonImpl getLiaison() {
        return liaison;
    }

    /**
     * Generates the name-representation of the connection instance
     */
    public String getName() {
        return this.address + ":" + this.port + "/" + this.serverName;
    }

    /**
     * Closes connection
     */
    public void finish() {
        try {
            jmxc.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

}

