/*
 * Copyright (c) 2012 Imre Fazekas.
 *  All rights reserved.
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *
 *  Redistributions of source code must retain the above copyright notice, this
 *  list of conditions and the following disclaimer.
 *
 *  Redistributions in binary form must reproduce the above copyright notice,
 *  this list of conditions and the following disclaimer in the documentation
 *  and/or other materials provided with the distribution.
 *  Neither the name of the Brillien nor the names of its
 *  terms and concepts may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 *  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 */

package com.vii.brillien.ignition.transport.websocket;

import com.vii.brillien.ignition.BrillienContext;
import com.vii.brillien.ignition.transport.BrillienCommunication;
import com.vii.brillien.kernel.BrillienException;
import com.vii.brillien.kernel.axiom.transport.Communication;
import com.vii.streamline.services.IOServices;
import com.vii.streamline.services.error.StreamLineException;
import com.vii.streamline.services.json.JsonServices;
import com.vii.streamline.websocket.WebSocketClient;
import org.svenson.matcher.PathMatcher;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;

/**
 * Specialized easy-to-use WebSocketClient
 */
public class PuritySocketClient  extends WebSocketClient {

    /**
     * Tells whether compression should be used
     */
    protected boolean                                                               compress;

    /**
     * Timeout of the answer-waiting process
     */
    protected long                                                                  timeout;

    /**
     * Internal structure to mainting communication flows
     */
    protected ConcurrentHashMap<String, LinkedBlockingQueue<BrillienCommunication>> responseQueues;

    {
        timeout         = 2000;
        responseQueues  = new ConcurrentHashMap<String, LinkedBlockingQueue<BrillienCommunication>>();
    }

    public PuritySocketClient(String wsUrl) throws StreamLineException {
        super(wsUrl);
    }

    public PuritySocketClient(String wsUrl, String keystoreFile, String keystorePassword, String keystoreType) throws StreamLineException {
        super(wsUrl, keystoreFile, keystorePassword, keystoreType);
    }

    public PuritySocketClient(String wsUrl, String keystoreFile, String keystorePassword, String keystoreType, String truststoreFile, String truststorePassword, String truststoreType) throws StreamLineException {
        super(wsUrl, keystoreFile, keystorePassword, keystoreType, truststoreFile, truststorePassword, truststoreType);
    }

    /**
     * Getter method for field compress
     */
    public boolean isCompress() {
        return compress;
    }

    /**
     * Setter method for field compress
     */
    public void setCompress(boolean compress) {
        this.compress = compress;
    }

    /**
     * Getter method for field timeout
     */
    public long getTimeout() {
        return timeout;
    }

    /**
     * Setter method for field timeout
     */
    public void setTimeout(long timeout) {
        this.timeout = timeout;
    }

    /**
     * Logs message
     */
    protected void logMessage( BrillienCommunication message ){
        BrillienContext.basicLog( Level.FINE, "Message RECEIVED:: " + message);
    }

    /**
     * Logs error message
     */
    protected void logErrorMessage( BrillienCommunication message ){
        BrillienContext.basicLog( Level.FINE, "ERROR RECEIVED:: " + message );
    }


    @Override
    public void onMessage(final String request) {
        try {
            BrillienCommunication msg = JsonServices.parseJSON(request, BrillienCommunication.class);
            logMessage( msg );
            if( msg.getMessageType() == Communication.TYPE_RESULT || msg.getMessageType() == Communication.TYPE_ERROR ){
                if( responseQueues.containsKey( msg.getOriginalId() ) ){
                    try{
                        responseQueues.get( msg.getOriginalId() ).offer( msg );
                    } catch (Exception e){ }
                }
            } else
                System.out.println("EVENT:::" + msg);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    @Override
    public void onMessage(byte[] bytes) {
        String request;
        try {
            request = IOServices.gzipDecompress(bytes);

            onMessage( request );
        } catch (StreamLineException e) {
            e.printStackTrace();
        }
    }

    protected BrillienCommunication  sendMessage(BrillienCommunication comm) throws BrillienException, StreamLineException {
        final LinkedBlockingQueue<BrillienCommunication> queue = new LinkedBlockingQueue<BrillienCommunication>(1);

        if( comm.getMessageType() == Communication.TYPE_GET )
            responseQueues.put( comm.getId(), queue );

        getSocket().sendTextMessage( JsonServices.toJSON( comm ) );

        if( comm.getMessageType() == Communication.TYPE_GET ){
            try {
                BrillienCommunication response = queue.poll(timeout, TimeUnit.MILLISECONDS);

                if (response == null) {
                    throw new BrillienException( "No response from server for communication: " + comm );
                }
                else if (response.getMessageType() == Communication.TYPE_ERROR) {
                    logErrorMessage( response );
                    throw new BrillienException( response.<String>getResponse() );
                }

                return response;
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
            throw new BrillienException( StreamLineException.ERROR_BROKEN_EXTERNAL, "Cannot get an answer from the server." );
        }

        return null;
    }

    /**
     * Sends the given communication and waits for an answer which will be mapped following the given typemappings
     */
    protected <T> T sendMessage(BrillienCommunication comm, Map<PathMatcher,Class> typehints, Class<T> responseType) throws BrillienException, StreamLineException {
        BrillienCommunication response = sendMessage( comm );

        return response == null
                ? null
                : typehints == null
                ? response.<T>acquireResponse( responseType )
                : response.<T>acquireResponse( typehints, responseType );
    }

    /**
     * Sends the given communication and waits for an answer which will be mapped following the given typemappings
     */
    protected <T> T sendMessage(BrillienCommunication comm, Class<T> responseType, Map<String,Class> typehints) throws BrillienException, StreamLineException {
        BrillienCommunication response = sendMessage( comm );

        return  response == null
                ? null
                : responseType == null
                ? response.<T>acquireResponse( typehints )
                : typehints == null
                ? response.<T>acquireResponse( responseType )
                : response.<T>acquireResponse( responseType, typehints );
    }

    /**
     * Sends the given communication and waits for an answer which will be mapped following the given typemappings
     */
    protected <T> T sendMessage(BrillienCommunication comm, Class<T> responseType) throws BrillienException, StreamLineException {
        return this.<T>sendMessage( comm, responseType, null );
    }

    /**
     * Sends the given communication and waits for an answer which will be mapped following the given typemappings
     */
    protected <T> T sendMessage(BrillienCommunication comm, Map<String,Class> typehints) throws BrillienException, StreamLineException {
        return this.<T>sendMessage( comm, null, typehints );
    }

}

